from django.apps import apps
from django.conf import settings

from manticore.models import base


def is_search_index(model_or_obj):
    if type(model_or_obj) is not base.SearchIndexBase:
        model = model_or_obj.__class__
    else:
        model = model_or_obj
    return (issubclass(model, base.SearchIndex) or
            type(model) is base.SearchIndexBase)


# noinspection PyUnusedLocal
class ManticoreRouter:
    db_name = getattr(settings, 'MANTICORE_DATABASE_NAME', 'manticore')

    def db_for_read(self, model, **hints):
        if is_search_index(model):
            return self.db_name

    def db_for_write(self, model, **hints):
        if is_search_index(model):
            return self.db_name

    @staticmethod
    def allow_relation(obj1, obj2, **hints):
        # Joins between search index and anything else are not allowed
        if is_search_index(obj1) or is_search_index(obj2):
            return False
        # No decisions are made for unrelated models
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if not model_name:
            # we can't tell anything
            return None
        model = apps.get_model(app_label, model_name)
        if is_search_index(model):
            # indices are migrated only for manticore db
            return db == self.db_name
        if db == self.db_name:
            # in manticore only indices are migrated
            return False
        return None



import operator
from functools import reduce

from django.db.models import query
from django.db.models.sql import AND

from manticore.sphinxql.expressions import T, Match, F
from manticore.sphinxql.base import SphinxQLCombinable, SphinxQLNode


class SearchQuerySet(query.QuerySet):
    def match(self, *args, **kwargs):
        """ Very large usage description here."""
        qs: SearchQuerySet = self._clone()
        expression = self._build_match_expression(*args, **kwargs)
        qs._match_expression().add(expression)
        return qs

    @staticmethod
    def _build_match_expression(*args, **kwargs):
        """ Transforms *args, **kwargs to SphinxQL DSL."""
        terms = []
        for term in args:
            if isinstance(term, str):
                terms.append(T(term))
            elif isinstance(term, (SphinxQLCombinable, SphinxQLNode)):
                terms.append(term)
            else:
                raise TypeError(term)
        terms.extend(map(lambda pair: F(*pair), kwargs.items()))
        return reduce(operator.and_, terms)

    def _match_expression(self) -> Match:
        """ Gets or adds Match expression to where clause."""
        where = self.query.where
        if where.connector != AND:
            raise ValueError(f"MATCH can't be used with {where.connector}")

        for node in where.children:
            if isinstance(node, Match):
                match = node
                break
        else:
            match = Match()
            where.add(match, AND)

        return match



from .base import SearchIndex
from .manager import SearchManager
from .query import SearchQuerySet
from .fields import *

__all__ = fields.__all__ + ['SearchIndex', 'SearchManager', 'SearchQuerySet']



from django.db import ProgrammingError
from django.db.backends.mysql import schema
from django.db.models.fields import NOT_PROVIDED
from django.db.models.options import Options

from manticore.models import fields, base


class DatabaseSchemaEditor(schema.DatabaseSchemaEditor):

    def prepare_default(self, value):
        raise NotImplementedError()

    def create_model(self, model):
        """
        Marks table name for database prefix addition, adds stub rt_field for
        django internal tables,
        """
        # noinspection PyProtectedMember
        opts: Options = model._meta
        # mark table name to set database name prefix for created tables
        opts.db_table = self.connection.ops.mark_table_name(opts.db_table)
        opts.db_table.skip_cluster = True
        # manticore search does not allow creating tables without rt fields,
        # adding a stub field for it
        has_rt_index = False
        for f in opts.local_fields:
            if isinstance(f, fields.RTField):
                has_rt_index = True
                break
        if not has_rt_index:
            stub = fields.IndexedField(db_column='__stub__')
            try:
                stub.contribute_to_class(model, '_stub')
                super().create_model(model)
            finally:
                # removing stub to prevent fetching non-stored field
                opts.local_fields.remove(stub)
                opts.__dict__.pop('fields', None)
        else:
            super().create_model(model)
        cluster = self.connection.settings_dict.get('CLUSTER', '')
        if cluster:
            with self.connection._nodb_cursor() as c:
                cluster = self.connection.ops.quote_name(cluster)
                db_table = self.connection.ops.quote_name(opts.db_table)
                try:
                    c.execute(f'CREATE CLUSTER {cluster}')
                except ProgrammingError:
                    pass
                c.execute(f'ALTER CLUSTER {cluster} ADD {db_table}')
        opts.db_table.skip_cluster = False

    def skip_default(self, field):
        # manticore does not support defaults at all
        return True

    def add_field(self, model, field):
        """
        Adds new column to table and performs update of default value
        """
        # calling BaseDatabaseSchemaEditor.add_field to skip mysql
        # implementation
        super(schema.DatabaseSchemaEditor, self).add_field(model, field)

        if (self.skip_default(field) and
                field.default not in (None, NOT_PROVIDED)):
            effective_default = self.effective_default(field)
            # UPDATE needs WHERE clause
            # noinspection SqlNoDataSourceInspection,PyProtectedMember
            self.execute(
                'UPDATE %(table)s SET %(column)s = %%s WHERE `id` > %%s' % {
                    'table': self.quote_name(model._meta.db_table),
                    'column': self.quote_name(field.column),
                }, [effective_default, 0])

    def column_sql(self, model, field, include_default=False):
        if field.primary_key:
            # FIXME: create manual primary keys
            return None, ''
        # without null flag column sql is created with NOT NULL annotation,
        # which is not supported by manticore
        null, field.null = field.null, True
        try:
            return super().column_sql(model, field, include_default=False)
        finally:
            field.null = null

    def alter_unique_together(self, model, old_unique_together,
                              new_unique_together):
        # unique indexes not supported
        pass

    def table_sql(self, model):
        sql, params = super().table_sql(model)
        # noinspection PyProtectedMember
        opts = model._meta

        index_sql, index_params = [], []
        for k in base.INDEX_OPTIONS:
            try:
                v = getattr(opts, k)
            except AttributeError:
                continue
            index_sql.append(f'{k} = %s')
            index_params.append(str(v))
        if index_sql:
            sql += ' '.join(index_sql)
            params += index_params

        return sql, params



from django.db.models import base, options

from manticore.models.manager import SearchManager


__all__ = ['SearchIndex']


INDEX_OPTIONS = (
    'min_prefix_len',
    'regexp_filter',
    'blend_chars'
)

options.DEFAULT_NAMES += INDEX_OPTIONS


class SearchIndexBase(base.ModelBase):
    """ Search index metaclass used in ManticoreRouter."""


class SearchIndex(base.Model, metaclass=SearchIndexBase):
    """ Search index base model."""
    class Meta:
        abstract = True

    objects = SearchManager()



import json

from django.db import models

__all__ = ['BigMultiField', 'JSONField', 'MultiField', 'RTField']

from manticore.models import lookups


class RTField(models.TextField):
    """ Full-text search field (rt_field)."""

    def __init__(self, **kwargs):
        # null is not supported by manticore, '' is reasonable default
        kwargs['null'] = False
        kwargs['default'] = ''
        super().__init__(**kwargs)

    def db_type(self, connection):
        return 'text indexed stored'

    def get_internal_type(self):
        return "RTField"


class IndexedField(RTField):
    """ Non-stored search field which may still be used for querying."""

    def db_type(self, connection):
        return 'text indexed'


class JSONField(models.Field):
    """ JSON field (attr_json)."""
    def db_type(self, connection):
        return 'json'

    def get_internal_type(self):
        return "JSONField"

    def get_prep_value(self, value):
        if value is None:
            return ''
        return json.dumps(value)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def from_db_value(self, value, expression, connection):
        if value is None:
            return None
        return json.loads(value)


class MultiField(models.PositiveIntegerField):
    """ Multi-field (attr_multi). Contains a list of uint32."""
    def db_type(self, connection):
        return 'multi'

    def get_internal_type(self):
        return 'MultiField'

    def get_prep_value(self, value):
        return value

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def from_db_value(self, value, expression, connection):
        if not value:
            return []
        return list(map(int, value.split(',')))


MultiField.register_lookup(lookups.MultiExact)


class BigMultiField(models.BigIntegerField):
    """ Multi-field for big integer (attr_multi64). Contains a list of int64."""

    def db_type(self, connection):
        return 'multi64'

    def get_internal_type(self):
        return 'BigMultiField'

    def get_prep_value(self, value):
        return value

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def from_db_value(self, value, expression, connection):
        if not value:
            return []
        return list(map(int, value.split(',')))


BigMultiField.register_lookup(lookups.MultiExact)



from django.core.exceptions import EmptyResultSet
from django.db import models
from django.db.backends.mysql import compiler
from django.db.models import expressions, lookups
from django.db.models.sql.datastructures import BaseTable
from django.db.models.sql.where import ExtraWhere, AND

from manticore.models.lookups import InFunction
from manticore.models.sql.where import ManticoreWhereNode
from manticore.sphinxql.expressions import Match


class SphinxQLCompiler(compiler.SQLCompiler):

    def compile(self, node):
        if isinstance(node, expressions.Col):
            # `table_name`.`column_name` is not supported
            return self.__compile_col(node)
        if isinstance(node, lookups.In):
            # col IN (values list) not supported, transforming to function call
            return self._compile_in(node)
        if isinstance(node, BaseTable):
            # prefix table name with database name
            return self.__compile_table(node)
        return super().compile(node)

    def as_sql(self, with_limits=True, with_col_aliases=False):
        self.__maybe_set_limits(with_limits)
        try:
            self.__maybe_move_where()
        except EmptyResultSet:
            # Where node has compiled to always false condition, but we need
            # to call super().as_sql again for some django internal side effects
            pass

        return super().as_sql(with_limits, with_col_aliases)

    def __compile_col(self, node: expressions.Col):
        qn = self.quote_name_unless_alias
        return qn(node.target.column), ()

    def __compile_table(self, node: BaseTable):
        qn = self.connection.ops.quote_name
        table_name = self.connection.ops.mark_table_name(node.table_name)
        return qn(table_name), ()

    def __maybe_move_where(self):
        where = self.query.where.clone()
        match = None
        for node in where.children:
            if isinstance(node, Match):
                match = node
                break
        if match:
            where.children.remove(match)
        if not where:
            return
        sql, params = self.compile(where)
        extra_select = expressions.RawSQL(sql, params, models.BooleanField())
        extra_where = ExtraWhere(['__where__ = %s'], (True,))
        where = ManticoreWhereNode()
        where.add(extra_where, AND)
        if match:
            where.add(match, AND)

        # All filtering conditions are now evaluated as __where__ in select
        # clause, so we need to check only that it is true
        self.query.add_annotation(extra_select, '__where__')
        self.query.where = where

    def __maybe_set_limits(self, with_limits):
        if with_limits and not self.query.low_mark:
            # by default 20 items are returned, setting to max value
            self.query.set_limits(high=2 ** 31 - 1)

    def _compile_in(self, node: lookups.In):
        lookup = InFunction(node.lhs, node.rhs)
        return lookup.as_sql(self, self.connection)



"""
This module contains base classes for SphinxQL domain language.
"""

from typing import Tuple, Iterable, Any


class SphinxQLNode:
    """
    SphinxQL non-leaf node.

    SphinxQLNode implements a node that combines child nodes with same connector.
    """
    AND = "&"
    OR = "|"
    MAYBE = "MAYBE"  # Lazy OR

    connector = AND

    def __init__(self, *expressions, connector=AND):
        self.expressions = list(expressions)
        self.connector = connector

    def as_sphinxql(self):
        sphinxql = []
        params = []
        for expr in self.expressions:
            s, p = expr.as_sphinxql()
            sphinxql.append(s)
            params.extend(p)
        connector = f' {self.connector} '
        if len(self.expressions) > 1:
            return f'({connector.join(sphinxql)})', params
        return connector.join(sphinxql), params

    def _combine(self, other, connector, reverse):
        if not hasattr(other, 'as_sphinxql'):
            # Graph can contain only nodes that render SphinxQL expressions.
            raise TypeError(other)

        if self.connector == connector:
            # With same connector we can use same predescence level.

            # Unpacking "other hand" expressions to list from SphinxQLNode
            expressions = list(getattr(other, 'expressions', [other]))
            if reverse:
                expressions.extend(self.expressions)
            else:
                expressions = self.expressions + expressions
            return self.__class__(*expressions, connector=connector)
        # Connector differs, add new level of predescence to graph.
        if reverse:
            return self.__class__(other, self, connector=connector)
        return self.__class__(self, other, connector=connector)

    def __and__(self, other):
        return self._combine(other, self.AND, False)

    def __or__(self, other):
        return self._combine(other, self.OR, False)

    def __rand__(self, other):
        return self._combine(other, self.AND, True)

    def __ror__(self, other):
        return self._combine(other, self.OR, True)


class SphinxQLCombinable:
    """
    Leaf node class for SphinxQL graph.

    SphinxQLCombinable is a base class for terms and unary operators in SphinxQL.

    It does not provide __init__ signature to allow any signature for child classes.
    """
    node_class = SphinxQLNode

    def __and__(self, other):
        return self._connect(other, self.node_class.AND, False)

    def __rand__(self, other):
        return self._connect(other, self.node_class.AND, True)

    def __or__(self, other):
        return self._connect(other, self.node_class.OR, False)

    def __ror__(self, other):
        return self._connect(other, self.node_class.OR, True)

    def _connect(self, other, connector, reverse):
        parts = [other, self] if reverse else [self, other]
        return self.node_class(*parts, connector=connector)

    def as_sphinxql(self) -> Tuple[str, Iterable[Any]]:
        raise NotImplementedError

    def __repr__(self):
        sql, params = self.as_sphinxql()
        return f'{self.__class__.__name__}: {sql % tuple(params)}'


ESCAPE = str.maketrans({k: rf"\{k}" for k in r'''!"$'()-/<@^|~'''})


def escape(s: str):
    s = s.replace('\\', '\\\\')
    s = s.translate(ESCAPE)
    return s



from django.db.backends.mysql import compiler
from django.db.models import sql, lookups

from manticore.models import RTField, JSONField
from manticore.models.sql.compiler import SphinxQLCompiler


class SQLCompiler(SphinxQLCompiler):
    """ interface for DatabaseOperations.compiler_module """
    pass


class SQLInsertCompiler(compiler.SQLInsertCompiler, SphinxQLCompiler):

    def execute_sql(self, returning_fields=None):
        """
        Marks table name for database name prefix addition.
        """
        # noinspection PyProtectedMember
        opts = self.query.model._meta
        # marking db_table attribute to add database name prefix in quote_name
        opts.db_table = self.connection.ops.mark_table_name(opts.db_table)
        return super().execute_sql(returning_fields)


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SphinxQLCompiler):

    def as_sql(self):
        """
        Performs REPLACE query instead of UPDATE in case of
        rt_field and attr_json updates.
        """
        if self.__is_index_field_update():
            pk = self.__get_primary_key_value()
            if pk is None:
                raise NotImplementedError(
                    "Updating indexed fields is supported only by pk value")
            # rt fields and attr_json can't be updated with UPDATE syntax;
            # if we are performing SearchIndex.save(), we can perform REPLACE
            return self.__as_replace(pk)

        return super().as_sql()

    def __get_primary_key_value(self):
        """
        :returns: pk value from WHERE clause which looks like "WHERE id = %s"
        """
        if not self.query.where:
            return
        if len(self.query.where.children) != 1:
            return
        node = self.query.where.children[0]
        if node.lookup_name != 'exact':
            return
        if not node.lhs.field.primary_key:
            return
        return node.rhs

    def __as_replace(self, pk):
        """
        Executes REPLACE query to update rt_field and attr_json fields by
        given pk value.
        """
        query = sql.InsertQuery(self.query.model)
        # noinspection PyProtectedMember
        opts = self.query.model._meta
        fields = [opts.pk]
        obj = self.query.model()
        setattr(obj, 'pk', pk)
        for field, _, value in self.query.values:
            fields.append(field)
            setattr(obj, field.attname, value)
        if set(fields) ^ set(opts.local_fields):
            # If you call REPLACE omitting some attributes, manticore will
            # store default empty values (0, '', [] etc...)
            raise NotImplementedError(
                "REPLACE query from save() with update_fields not supported")
        query.insert_values(fields, [obj], raw=False)
        insert_compiler = query.get_compiler(self.using, self.connection)
        sqls = insert_compiler.as_sql()
        insert_sql, params = sqls[0]

        # INSERT -> REPLACE
        if not insert_sql.startswith('INSERT '):
            raise ValueError(insert_sql)
        insert_sql = 'REPLACE' + insert_sql[6:]

        return insert_sql, params

    def __is_index_field_update(self):
        """ Checks whether update query contains indexed fields updates."""
        for field, _, _ in self.query.values:
            if isinstance(field, (RTField, JSONField)):
                return True
        return False


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SphinxQLCompiler):
    def _compile_in(self, node: lookups.In):
        """
        Formats "a IN values" expression instead of widely used in manticore
        "IN(a, values)" function
        """
        # DELETE supports WHERE id IN (values_list) instead of IN-function
        return node.as_sql(self, self.connection)

    def _as_sql(self, query):
        # mark table name to be quoted with test database prefix, see
        # ManticoreOperations.quote_name
        query.base_table = self.connection.ops.mark_table_name(query.base_table)
        return super()._as_sql(query)



from datetime import datetime

from django.db import ProgrammingError
from django.db.backends.base.introspection import TableInfo
from django.db.backends.mysql import base
from django.utils import timezone
from django.utils.functional import cached_property

from manticore.backend.schema import DatabaseSchemaEditor


class TableName(str):
    """
    Table name marker for proper database name prefix addition.
    """
    # table name marker used in quote_name for adding database name prefix
    # to table names
    is_table_name = True


class DatabaseWrapper(base.DatabaseWrapper):
    data_types = {
        **base.DatabaseWrapper.data_types,
        # overriding model fields mapping to manticore attributes
        'CharField': 'string',  # attr_string instead of varchar
        'TextField': 'string',  # attr_string instead of text, which is rt_field
        # closest equivalent for datetime field in manticore is attr_timestamp
        'DateTimeField': 'timestamp',
        'FloatField': 'float',  # attr_float instead of double
    }

    SchemaEditorClass = DatabaseSchemaEditor

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.creation = ManticoreCreation(self)
        self.features = ManticoreFeatures(self)
        self.introspection = ManticoreIntrospection(self)
        self.ops = ManticoreOperations(self)
        self.validation = ManticoreValidation(self)

    @cached_property
    def mysql_server_info(self):
        # mysql uses 'SELECT VERSION()', not supported
        with self.temporary_connection():
            return self.connection.get_server_info()


class ManticoreFeatures(base.DatabaseFeatures):
    # mysql detects this querying SELECT @@SQL_AUTO_IS_NULL, not supported
    is_sql_auto_is_null_enabled = False
    # column definition NULL is not supported
    implied_column_null = True
    # FIXME add fake transaction support
    # manticore supports over-isolated transaction, which means that updates
    # are not visible in same transaction. Transaction support is disabled
    # mostly for running tests.
    supports_transactions = False
    # django tries to check foreign key constraints in tests
    can_rollback_ddl = False
    # select for update not supported
    has_select_for_update = False
    # savepoints not supported
    uses_savepoints = False
    # enables returning primary keys from LAST_INSERT_ID() in bulk_create
    can_return_rows_from_bulk_insert = True
    can_return_columns_from_insert = True


class ManticoreIntrospection(base.DatabaseIntrospection):

    def get_table_list(self, cursor):
        """Return a list of table and view names in the current database."""
        # mysql uses 'SHOW FULL TABLES', not supported
        cursor.execute("SHOW TABLES")
        result = []
        db_name = self.connection.ops.db_name
        database_prefix = f'{db_name}__' if db_name else None
        # manticore annotates real-time indices with "rt" type
        table_types = {
            'rt': 't'
        }
        for row in cursor.fetchall():
            name = row[0]
            table_type = row[1]
            # multi-database support is implemented with table prefixes,
            # that should be removed when collecting back table information
            if database_prefix and name.startswith(database_prefix):
                # removing database prefix for tables
                name = name[len(database_prefix):]
            elif database_prefix is not None:
                # We don't need tables not from current db
                continue
            result.append(TableInfo(name, table_types.get(table_type)))
        # FIXME: support on-disk indices as views
        return result

    def get_storage_engine(self, cursor, table_name):
        # 'SELECT * FROM information_schema.tables', not supported
        # noinspection PyProtectedMember
        return 'RT'


class ManticoreValidation(base.DatabaseValidation):
    def _check_sql_mode(self, **kwargs):
        # mysql uses 'SELECT @@sql_mode', not supported
        return []


class ManticoreOperations(base.DatabaseOperations):
    compiler_module = 'manticore.backend.compiler'

    @property
    def db_name(self):
        return self.connection.settings_dict.get('NAME', '')

    @property
    def cluster_name(self):
        return self.connection.settings_dict.get('CLUSTER', '')

    def quote_name(self, name):
        """ Table names are prefixed with database name."""
        # Multi-database support is implemented with database name prefixes
        # for tables, but this method is used also for quoting other identifiers
        # like database name or column names. To distinguish table names
        # `mark_table_name` method is used to add table name mark for `name`
        # argument.
        is_table_name = getattr(name, 'is_table_name', False)
        skip_cluster = getattr(name, 'skip_cluster', False)
        if is_table_name and self.db_name:
            name = f'{self.db_name}__{name}'
        if is_table_name and self.cluster_name and not skip_cluster:
            cluster = super().quote_name(self.cluster_name)
            name = super().quote_name(name)
            return f'{cluster}:{name}'
        return super().quote_name(name)

    @staticmethod
    def mark_table_name(name):
        """
        Marks table name with is_table_name flag for correct addition of
        database name prefix
        """
        if isinstance(name, TableName):
            return name
        return TableName(name)

    def insert_statement(self, ignore_conflicts=False):
        return 'REPLACE INTO' if ignore_conflicts else 'INSERT INTO'

    def adapt_datetimefield_value(self, value):
        """ Converts datetime value to unix timestamp."""
        if isinstance(value, datetime):
            # closest manticore datetime equivalent is attr_timestamp
            return int(timezone.utc.normalize(value).timestamp())
        return super().adapt_datetimefield_value(value)

    def convert_datetimefield_value(self, value, expression, connection):
        """ Convert unix timestamp values to datetime."""
        if isinstance(value, int):
            # closest manticore datetime equivalent is attr_timestamp, which
            # stores datetime as unix utc timestamp
            value = datetime.utcfromtimestamp(value)
        return super().convert_datetimefield_value(
            value, expression, connection)

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        """ Implement flushing manticore database as truncating all tables."""
        if tables:
            sql = []
            for table in tables:
                table = self.mark_table_name(table)
                sql.append('%s %s;' % (
                    style.SQL_KEYWORD('TRUNCATE RTINDEX'),
                    style.SQL_FIELD(self.quote_name(table)),
                ))
            return sql
        else:
            return []

    def fetch_returned_insert_rows(self, cursor):
        cursor.execute("SELECT LAST_INSERT_ID()")
        row = cursor.fetchone()
        result = []
        for value in map(int, row[0].split(',')):
            result.append([value])
        return result

    def return_insert_columns(self, fields):
        return '', []

    def fetch_returned_insert_columns(self, cursor, returning_params):
        # return tuple to avoid extending inserted_rows list in _batched_insert
        return cursor.lastrowid,


class ManticoreCreation(base.DatabaseCreation):

    def _execute_create_test_db(self, cursor, parameters, keepdb=False):
        # manticore does not support multiple databases, skipping
        # CREATE DATABASE command
        pass

    def _destroy_test_db(self, test_database_name, verbosity):
        # manticore does not support multiple databases, skipping
        # DROP DATABASE command.
        # noinspection PyProtectedMember
        with self.connection._nodb_cursor() as c:
            cluster = self.connection.settings_dict.get('CLUSTER', '')
            if cluster:
                cluster = self.connection.ops.quote_name(cluster)
                try:
                    c.execute(f"DELETE CLUSTER {cluster}")
                except ProgrammingError:
                    pass
            # manticore does not support destroying databases, instead we
            # drop every table with corresponding prefix
            for table in self.connection.introspection.get_table_list(c):
                c.execute(f"DROP TABLE {test_database_name}__{table.name}")

    def _clone_db(self, source_database_name, target_database_name):
        # copying tables with source prefix to target prefix
        # noinspection PyProtectedMember
        with self.connection._nodb_cursor() as c:
            cluster = self.connection.settings_dict.get('CLUSTER', '')
            if cluster:
                cluster = self.connection.ops.quote_name(cluster)
                try:
                    c.execute(f"CREATE CLUSTER {cluster}")
                except ProgrammingError:
                    pass
            for table in self.connection.introspection.get_table_list(c):
                c.execute(f"CREATE TABLE {target_database_name}__{table.name} "
                          f"LIKE {source_database_name}__{table.name}")
                if cluster:
                    c.execute(f"ALTER CLUSTER {cluster} ADD {target_database_name}__{table.name}")



from django.db.models import manager

from manticore.models.query import SearchQuerySet


class SearchManager(manager.BaseManager.from_queryset(SearchQuerySet)):
    """ Manager for search indices."""
    # Copies SearchQuerySet.match() method to Manager



from django.db.models.sql.where import WhereNode


class ManticoreWhereNode(WhereNode):
    # Manticore does not allow parentheses on root where clause
    def as_sql(self, compiler, connection):
        sql, params = super().as_sql(compiler, connection)
        # remove one parentheses level
        if sql[0] == '(' and sql[-1] == ')':
            sql = sql[1:-1]
        return sql, params



import os
import re
import subprocess
from setuptools import setup, find_packages  # type: ignore
from pathlib import Path

with open('README.md') as f:
    long_description = f.read()

version_re = re.compile('^Version: (.+)$', re.M)
package_name = 'django-manticore'


def get_version():
    """
    Reads version from git status or PKG-INFO

    https://gist.github.com/pwithnall/7bc5f320b3bdf418265a
    """
    d: Path = Path(__file__).absolute().parent
    git_dir = d.joinpath('.git')
    if git_dir.is_dir():
        # Get the version using "git describe".
        cmd = 'git describe --tags --match [0-9]*'.split()
        try:
            version = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            return None

        # PEP 386 compatibility
        if '-' in version:
            version = '.post'.join(version.split('-')[:2])

        # Don't declare a version "dirty" merely because a time stamp has
        # changed. If it is dirty, append a ".dev1" suffix to indicate a
        # development revision after the release.
        with open(os.devnull, 'w') as fd_devnull:
            subprocess.call(['git', 'status'],
                            stdout=fd_devnull, stderr=fd_devnull)

        cmd = 'git diff-index --name-only HEAD'.split()
        try:
            dirty = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            return None

        if dirty != '':
            version += '.dev1'
    else:
        # Extract the version from the PKG-INFO file.
        try:
            with open('PKG-INFO') as v:
                version = version_re.search(v.read()).group(1)
        except FileNotFoundError:
            version = None

    return version


setup(
    name=package_name,
    version=get_version() or 'dev',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(where='.', exclude=['testproject', 'testproject.*']),
    url='https://github.com/just-work/django-manticore',
    license='MIT',
    author='Sergey Tikhonov',
    author_email='zimbler@gmail.com',
    description='Manticoresearch database backend for Django',
    install_requires=[
        'Django>=3.1,<3.3',
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Distributed Computing',
    ]
)



"""
This module contains public interface for SphinxQL domain language.
"""

from typing import Optional, Union

from manticore.sphinxql.base import SphinxQLCombinable, SphinxQLNode, escape

__all__ = [
    'F',
    'Match',
    'P',
    'T',
]


class TextNode(SphinxQLNode):
    """ Node allowing operations with text terms."""


class T(SphinxQLCombinable):
    """
    SphinxQL text term.

    Supports:

    - prefix search (text*)
    - start of text search (^text)
    - negate

    """
    node_class = TextNode

    def __init__(self, term: str, negate=False, exact=False, prefix=False, start=False):
        """
        Initializes search term node:

        >>> T("any of word matches")
        T: (any of word matches)
        >>> T("negated", negate=True)
        T: !(negated)


        """
        if not isinstance(term, str):
            raise TypeError("term is not string")
        self.term = term
        self.negate = negate
        self.exact = exact
        self.prefix = prefix
        self.start = start

    def __invert__(self):
        return self.__class__(self.term, negate=not self.negate,
                              exact=self.exact)

    def as_sphinxql(self):
        p = ''
        if self.exact:
            p = '='
        elif self.start:
            p = '^'
        s = ''
        if self.prefix:
            s = '*'
        sql = f'!({p}%s{s})' if self.negate else f'({p}%s{s})'
        return sql, [self.term]


class P(T):
    """
    SphinxQL quoted text node (phrase).
    >>> P("phrase search")
    P: ("phrase search")
    >>> ~P("exclude phrase")
    P: !("exclude phrase")
    >>> P("2 of m words match", quorum=2)
    P: ("2 of m words match"/2)
    >>> P("~30% of words match", quorum=0.3)
    P: ("~30% of words match"/0.3)
    >>> P("distance between words", proximity=4)
    P: ("distance between words"~4)
    >>> ~P("exact word forms", exact=True)
    P: !(="exact word forms")
    """

    def __init__(self, term: str, negate=False, exact=False,
                 proximity: Optional[int] = None,
                 quorum: Union[None, int, float] = None):
        super().__init__(term, negate, exact)
        if proximity is not None:
            if not isinstance(proximity, int):
                raise TypeError("proximity must be int")
        self.proximity = proximity
        if quorum is not None:
            if not isinstance(quorum, (int, float)):
                raise TypeError("quorum must be int or float")
        self.quorum = quorum

    def __invert__(self):
        return self.__class__(self.term, negate=not self.negate,
                              exact=self.exact, proximity=self.proximity,
                              quorum=self.quorum)

    def as_sphinxql(self):
        if isinstance(self.quorum, int):
            m = f'/{self.quorum}'
        elif isinstance(self.quorum, float):
            m = f'/{self.quorum:0.7f}'.rstrip('0')
        elif self.proximity:
            m = f'~{self.proximity}'
        else:
            m = ''
        p = '=' if self.exact else ''
        sql = f'!({p}"%s"{m})' if self.negate else f'({p}"%s"{m})'
        return sql, [self.term]


class F(SphinxQLCombinable):
    """
    Field search operator

    Supports:
    - @field text -

    """

    def __init__(self, *args, exclude=False, **kwargs):
        """
        Constructs field search expression
        :param args: field names and search expression (last args item)
        :param exclude: exclude fields from search
        :param kwargs: shortcut for single field name -> search expression value

        >>> F(first_field="text")
        F: (@first_field (text))
        >>> F('first_field', 'other_field', T("term"))
        F: (@(first_field,other_field) (term))
        >>> F(other_field=(T("text") & ~T("other")), exclude=True)
        F: (@!other_field (text) & (other))
        """
        if args and kwargs:
            raise ValueError("Don't pass args and kwargs simultaneously")

        if args:
            if len(args) < 2:
                raise ValueError(
                    "pass at least one field name and single search expression")
            *field_names, expression = args
        elif kwargs:
            if len(kwargs) > 1:
                raise ValueError("Only single value in kwargs is supported")
            tup = kwargs.items().__iter__().__next__()
            field_names = (tup[0],)
            expression = tup[1]
        else:
            raise ValueError("Pass args or kwargs")

        self.fields = field_names
        if isinstance(expression, str):
            self.expression = T(expression)
        elif isinstance(expression, T):
            self.expression = expression
        elif isinstance(expression, TextNode):
            self.expression = expression
        else:
            raise TypeError("unsupported expression for F")
        self.exclude = exclude

    def as_sphinxql(self):
        if len(self.fields) == 1:
            fields = self.fields[0]
        else:
            fields = f"({','.join(self.fields)})"
        prefix = '@!' if self.exclude else '@'

        expr, params = self.expression.as_sphinxql()
        return f'({prefix}{fields} {expr})', params


class Match(SphinxQLNode):
    """ Root match node that renders MATCH(...) """
    # duck typing for Django ORM
    contains_aggregate = False

    def add(self, *expressions):
        """ Inplace extending match query."""
        self.expressions.extend(expressions)

    # noinspection PyUnusedLocal
    def as_sql(self, compiler, connection):
        sphinxql, params = self.as_sphinxql()
        expression = sphinxql % tuple(map(escape, params))
        return f"MATCH(%s)", [expression]



from django.core.exceptions import EmptyResultSet
from django.db.models import lookups
from django.utils.datastructures import OrderedSet


class MultiExact(lookups.Exact):
    """ Implements equality comparison for multi fields with IN-function."""

    def as_sql(self, compiler, connection):
        return InFunction(self.lhs, [self.rhs]).as_sql(compiler, connection)


class InFunction(lookups.In):
    """ Implements function IN(needle, haystack...)."""

    def as_sql(self, compiler, connection):
        """
        Formats IN function call
        """
        lhs_sql, params = self.process_lhs(compiler, connection)
        rhs_sql, rhs_params = self.process_rhs(compiler, connection)
        params.extend(rhs_params)
        return 'IN(%s, %s)' % (lhs_sql, rhs_sql), params

    def process_rhs(self, compiler, connection):
        """
        Forms placeholder list without parentheses like "%s %s %s", for use
        with function call IN(%s, %s, %s, %s).
        """
        if self.rhs_is_direct_value():
            try:
                rhs = OrderedSet(self.rhs)
            except TypeError:  # Unhashable items in self.rhs
                rhs = self.rhs

            if not rhs:
                raise EmptyResultSet

            # rhs should be an iterable; use batch_process_rhs() to
            # prepare/transform those values.
            sqls, sqls_params = self.batch_process_rhs(
                compiler, connection, rhs)
            placeholder = ', '.join(sqls)
            return placeholder, sqls_params
        else:
            # subqueries not supported
            raise NotImplementedError()
